import { jsx, jsxs } from 'react/jsx-runtime';
import { memo, forwardRef, useMemo, useState, useRef, useCallback, useEffect, useLayoutEffect, useImperativeHandle } from 'react';
import ResizeObserver from 'resize-observer-polyfill';
import { parse, LineType } from 'clrc';

const LINE_CLASSNAME = 'react-lrc-line';

const baseStyle = {
    outline: 'none',
    boxSizing: 'border-box',
    scrollBehavior: 'smooth',
    overflow: 'auto',
};
function Root({ style, ...props }, ref) {
    const combinedStyle = useMemo(() => ({
        ...baseStyle,
        ...style,
    }), [style]);
    return jsx("div", { ...props, style: combinedStyle, ref: ref });
}
var Root$1 = memo(forwardRef(Root));

const style = {
    height: '50%',
};
function Space() {
    return jsx("div", { style: style });
}
var Space$1 = memo(Space);

function useLineIndex(lines, currentMillisecond) {
    let index = 0;
    for (const { length } = lines; index <= length; index += 1) {
        const line = lines[index];
        if (!line || line.startMillisecond > currentMillisecond) {
            break;
        }
    }
    return index - 1;
}

const DEFAULT_PROPS = {
    currentMillisecond: -1,
    verticalSpace: false,
    recoverAutoScrollInterval: 5000,
    recoverAutoScrollSingal: false,
};

function throttle(f, wait = 100) {
    let lastEmitTimestamp = 0;
    return (...params) => {
        const now = Date.now();
        if (now - lastEmitTimestamp >= wait) {
            lastEmitTimestamp = now;
            return f(...params);
        }
    };
}

const SCROLLABLE_KEYS = [' ', 'ArrowUp', 'ArrowDown'];
/**
 * detect user scroll
 * 1. wheel
 * 2. keyboard
 * 3. drag scrollbar
 * @author mebtte<hi@mebtte.com>
 */
var useAutoScroll = ({ recoverAutoScrollInterval, recoverAutoScrollSingal, }) => {
    const [autoScroll, setAutoScroll] = useState(true);
    const timerRef = useRef();
    const handleUserScroll = useCallback(() => {
        globalThis.clearTimeout(timerRef.current);
        setAutoScroll(false);
        timerRef.current = globalThis.setTimeout(() => setAutoScroll(true), recoverAutoScrollInterval);
    }, [recoverAutoScrollInterval]);
    const mouseDownRef = useRef(false);
    const onMouseDown = useCallback(() => {
        mouseDownRef.current = true;
    }, []);
    const onMouseUp = useCallback(() => {
        mouseDownRef.current = false;
    }, []);
    const onMouseMove = useMemo(() => throttle(() => {
        if (mouseDownRef.current) {
            handleUserScroll();
        }
    }), [handleUserScroll]);
    const onKeyDown = useMemo(() => throttle((event) => {
        if (SCROLLABLE_KEYS.includes(event.key)) {
            handleUserScroll();
        }
    }), [handleUserScroll]);
    const onWheel = useMemo(() => throttle(handleUserScroll), [handleUserScroll]);
    useEffect(() => {
        setAutoScroll(true);
    }, [recoverAutoScrollSingal]);
    /**
     * clear timer after unmount
     * @author mebtte<hi@mebtte.com>
     */
    useEffect(() => () => globalThis.clearTimeout(timerRef.current), []);
    return {
        autoScroll,
        onWheel,
        onKeyDown,
        onMouseDown,
        onMouseUp,
        onMouseMove,
    };
};

/**
 * stable useCallback
 * @author mebtte<hi@mebtte.com>
 */
function useEvent(callback) {
    const callbackRef = useRef();
    callbackRef.current = callback;
    const memoCallback = useCallback((...args) => callbackRef.current(...args), []);
    return memoCallback;
}

function debounce(f, wait = 300) {
    let timer;
    return (...params) => {
        globalThis.clearTimeout(timer);
        timer = globalThis.setTimeout(() => f(params), wait);
    };
}

/* eslint-disable no-param-reassign */
var useScrollAction = ({ root, autoScroll, lineIndex, lines, verticalSpace, }) => {
    const indexMapRef = useRef(new Map());
    const scrollToCurrentLine = useCallback(() => {
        if (!root) {
            return;
        }
        const indexMap = indexMapRef.current.get(lineIndex);
        if (indexMap) {
            root.scrollTop =
                indexMap.offsetTop - root.clientHeight * 0.5 + indexMap.height / 2;
        }
        else {
            root.scrollTop = 0;
        }
    }, [lineIndex, root]);
    useLayoutEffect(() => {
        if (root) {
            const caculateIndexMap = () => {
                const indexMap = new Map();
                const lrcLineNodeList = root.querySelectorAll(`.${LINE_CLASSNAME}`);
                for (let i = 0, { length } = lrcLineNodeList; i < length; i += 1) {
                    const lrcLineNode = lrcLineNodeList[i];
                    indexMap.set(i, {
                        height: lrcLineNode.clientHeight,
                        offsetTop: lrcLineNode.offsetTop,
                    });
                }
                indexMapRef.current = indexMap;
            };
            caculateIndexMap();
            const resizeDetector = new ResizeObserver(debounce(caculateIndexMap));
            resizeDetector.observe(root);
            return () => resizeDetector.disconnect();
        }
    }, [root, lines, verticalSpace]);
    useLayoutEffect(() => {
        if (autoScroll) {
            scrollToCurrentLine();
        }
    }, [autoScroll, scrollToCurrentLine, lines, verticalSpace]);
};

const space = jsx(Space$1, {});
function BaseLrc({ lines, lineRenderer, currentMillisecond = DEFAULT_PROPS.currentMillisecond, verticalSpace = DEFAULT_PROPS.verticalSpace, recoverAutoScrollInterval = DEFAULT_PROPS.recoverAutoScrollInterval, recoverAutoScrollSingal = DEFAULT_PROPS.recoverAutoScrollSingal, onLineUpdate, onWheel, onKeyDown, onMouseDown, onMouseUp, onMouseMove, ...props }, ref) {
    const rootRef = useRef(null);
    const lineIndex = useLineIndex(lines, currentMillisecond);
    const { autoScroll, onWheel: onLocalAutoScrollWheel, onKeyDown: onLocalAutoScrollKeyDown, onMouseDown: onLocalAutoScrollMouseDown, onMouseUp: onLocalAutoScrollMouseUp, onMouseMove: onLocalAutoScrollMove, } = useAutoScroll({
        recoverAutoScrollInterval,
        recoverAutoScrollSingal,
    });
    useScrollAction({
        root: rootRef.current,
        autoScroll,
        lineIndex,
        lines,
        verticalSpace,
    });
    useEffect(() => {
        if (onLineUpdate) {
            onLineUpdate({
                index: lineIndex,
                line: lines[lineIndex] ?? null,
            });
        }
    }, [onLineUpdate, lineIndex, lines]);
    useImperativeHandle(ref, () => rootRef.current);
    const onWheelWrapper = useEvent((event) => {
        onLocalAutoScrollWheel();
        return onWheel && onWheel(event);
    });
    const onKeyDownWrapper = useEvent((event) => {
        onLocalAutoScrollKeyDown(event);
        return onKeyDown && onKeyDown(event);
    });
    const onMouseDownWrapper = useEvent((event) => {
        onLocalAutoScrollMouseDown();
        return onMouseDown && onMouseDown(event);
    });
    const onMouseUpWrapper = useEvent((event) => {
        onLocalAutoScrollMouseUp();
        return onMouseUp && onMouseUp(event);
    });
    const onMouseMoveWrapper = useEvent((event) => {
        onLocalAutoScrollMove();
        return onMouseMove && onMouseMove(event);
    });
    const lineNodes = useMemo(() => lines.map((line, index) => (jsx("div", { className: LINE_CLASSNAME, children: lineRenderer({ index, active: lineIndex === index, line }) }, line.id))), [lineIndex, lineRenderer, lines]);
    return (jsxs(Root$1
    /** tabIndex make focusable and enable to add keyboard listener */
    , { 
        /** tabIndex make focusable and enable to add keyboard listener */
        tabIndex: -1, ...props, onWheel: onWheelWrapper, onKeyDown: onKeyDownWrapper, onMouseDown: onMouseDownWrapper, onMouseUp: onMouseUpWrapper, onMouseMove: onMouseMoveWrapper, ref: rootRef, children: [verticalSpace ? space : null, lineNodes, verticalSpace ? space : null] }));
}
var BaseLrc$1 = forwardRef(BaseLrc);

/**
 * get random string
 * @author mebtte<hi@mebtte.com>
 */
function getRandomString() {
    return Math.random().toString(36).substring(2, 8);
}

function useLrc(lrc) {
    const lines = useMemo(() => parse(lrc).filter((line) => line.type === LineType.LYRIC)
        .map((l) => ({
        id: getRandomString(),
        lineNumber: l.lineNumber,
        raw: l.raw,
        startMillisecond: l.startMillisecond,
        content: l.content,
    }))
        .sort((a, b) => a.startMillisecond - b.startMillisecond), [lrc]);
    return lines;
}

const Lrc$1 = ({ lrc, ...props }, ref) => {
    const lines = useLrc(lrc);
    return jsx(BaseLrc$1, { ...props, lines: lines, ref: ref });
};
var Lrc$2 = forwardRef(Lrc$1);

var useLrcs = (lrcs) => {
    const lines = useMemo(() => {
        const map = new Map();
        for (const lrc of lrcs) {
            const list = parse(lrc).filter((l) => l.type === LineType.LYRIC);
            for (const item of list) {
                const current = map.get(item.startMillisecond) || {
                    id: getRandomString(),
                    startMillisecond: item.startMillisecond,
                    children: [],
                };
                current.children.push({
                    id: getRandomString(),
                    lineNumber: item.lineNumber,
                    raw: item.raw,
                    content: item.content,
                });
                map.set(item.startMillisecond, current);
            }
        }
        return [...map.values()].sort((a, b) => a.startMillisecond - b.startMillisecond);
    }, [lrcs.length, ...lrcs]);
    return lines;
};

function Lrc({ lrcs, ...props }, ref) {
    const lines = useLrcs(lrcs);
    return jsx(BaseLrc$1, { ...props, lines: lines, ref: ref });
}
var MultipleLrc = forwardRef(Lrc);

function useRecoverAutoScrollImmediately() {
    const [signal, setSignal] = useState(false);
    const recoverAutoScrollImmediately = useCallback(() => setSignal((s) => !s), []);
    return { signal, recoverAutoScrollImmediately };
}

export { Lrc$2 as Lrc, MultipleLrc, useRecoverAutoScrollImmediately };
